@@ -5,7 +5,7 @@
    | This source file is subject to version 3.01 of the PHP license,      |
    | that is bundled with this package in the file LICENSE, and is        |
    | available through the world-wide-web at the following url:           |
-   | http://www.php.net/license/3_01.txt                                  |
+   | https://www.php.net/license/3_01.txt                                 |
    | If you did not receive a copy of the PHP license and are unable to   |
    | obtain it through the world-wide-web, please send a note to          |
    | license@php.net so we can mail you a copy immediately.               |
@@ -33,11 +33,6 @@
 
 #define SPL_HEAP_CORRUPTED       0x00000001
 
-#define SPL_PQUEUE_EXTR_MASK     0x00000003
-#define SPL_PQUEUE_EXTR_BOTH     0x00000003
-#define SPL_PQUEUE_EXTR_DATA     0x00000001
-#define SPL_PQUEUE_EXTR_PRIORITY 0x00000002
-
 zend_object_handlers spl_handler_SplHeap;
 zend_object_handlers spl_handler_SplPriorityQueue;
 
@@ -73,12 +68,6 @@ struct _spl_heap_object {
 	zend_object         std;
 };
 
-/* define an overloaded iterator structure */
-struct _spl_heap_it {
-	zend_user_iterator  intern;
-	int                 flags;
-};
-
 typedef struct _spl_pqueue_elem {
 	zval data;
 	zval priority;
@@ -97,7 +86,15 @@ static zend_always_inline void *spl_heap_elem(spl_ptr_heap *heap, size_t i) {
 
 static zend_always_inline void spl_heap_elem_copy(spl_ptr_heap *heap, void *to, void *from) {
 	assert(to != from);
-	memcpy(to, from, heap->elem_size);
+
+	/* Specialized for cases of heap and priority queue. With the size being
+	 * constant known at compile time the compiler can fully inline calls to memcpy. */
+	if (heap->elem_size == sizeof(spl_pqueue_elem)) {
+		memcpy(to, from, sizeof(spl_pqueue_elem));
+	} else {
+		ZEND_ASSERT(heap->elem_size == sizeof(zval));
+		memcpy(to, from, sizeof(zval));
+	}
 }
 
 static void spl_ptr_heap_zval_dtor(void *elem) { /* {{{ */
@@ -237,6 +234,22 @@ static int spl_ptr_pqueue_elem_cmp(void *x, void *y, zval *object) { /* {{{ */
 }
 /* }}} */
 
+/* Specialized comparator used when we are absolutely sure an instance of the
+ * not inherited SplPriorityQueue class contains only priorities as longs. This
+ * fact is tracked during insertion into the queue. */
+static int spl_ptr_pqueue_elem_cmp_long(void *x, void *y, zval *object) {
+	zend_long a = Z_LVAL(((spl_pqueue_elem*) x)->priority);
+	zend_long b = Z_LVAL(((spl_pqueue_elem*) y)->priority);
+	return a>b ? 1 : (a<b ? -1 : 0);
+}
+
+/* same as spl_ptr_pqueue_elem_cmp_long */
+static int spl_ptr_pqueue_elem_cmp_double(void *x, void *y, zval *object) {
+	double a = Z_DVAL(((spl_pqueue_elem*) x)->priority);
+	double b = Z_DVAL(((spl_pqueue_elem*) y)->priority);
+	return ZEND_NORMALIZE_BOOL(a - b);
+}
+
 static spl_ptr_heap *spl_ptr_heap_init(spl_ptr_heap_cmp_func cmp, spl_ptr_heap_ctor_func ctor, spl_ptr_heap_dtor_func dtor, size_t elem_size) /* {{{ */
 {
 	spl_ptr_heap *heap = emalloc(sizeof(spl_ptr_heap));
@@ -260,7 +273,7 @@ static void spl_ptr_heap_insert(spl_ptr_heap *heap, void *elem, void *cmp_userda
 	if (heap->count+1 > heap->max_size) {
 		size_t alloc_size = heap->max_size * heap->elem_size;
 		/* we need to allocate more memory */
-		heap->elements  = erealloc(heap->elements, 2 * alloc_size);
+		heap->elements  = safe_erealloc(heap->elements, 2, alloc_size, 0);
 		memset((char *) heap->elements + alloc_size, 0, alloc_size);
 		heap->max_size *= 2;
 	}
@@ -466,7 +479,7 @@ static zend_object *spl_heap_object_clone(zend_object *old_object) /* {{{ */
 }
 /* }}} */
 
-static int spl_heap_object_count_elements(zend_object *object, zend_long *count) /* {{{ */
+static zend_result spl_heap_object_count_elements(zend_object *object, zend_long *count) /* {{{ */
 {
 	spl_heap_object *intern = spl_heap_from_obj(object);
 
@@ -590,9 +603,9 @@ PHP_METHOD(SplHeap, insert)
 	zval *value;
 	spl_heap_object *intern;
 
-	if (zend_parse_parameters(ZEND_NUM_ARGS(), "z", &value) == FAILURE) {
-		RETURN_THROWS();
-	}
+	ZEND_PARSE_PARAMETERS_START(1, 1)
+		Z_PARAM_ZVAL(value);
+	ZEND_PARSE_PARAMETERS_END();
 
 	intern = Z_SPLHEAP_P(ZEND_THIS);
 
@@ -638,9 +651,10 @@ PHP_METHOD(SplPriorityQueue, insert)
 	spl_heap_object *intern;
 	spl_pqueue_elem elem;
 
-	if (zend_parse_parameters(ZEND_NUM_ARGS(), "zz", &data, &priority) == FAILURE) {
-		RETURN_THROWS();
-	}
+	ZEND_PARSE_PARAMETERS_START(2, 2)
+		Z_PARAM_ZVAL(data);
+		Z_PARAM_ZVAL(priority);
+	ZEND_PARSE_PARAMETERS_END();
 
 	intern = Z_SPLHEAP_P(ZEND_THIS);
 
@@ -652,6 +666,21 @@ PHP_METHOD(SplPriorityQueue, insert)
 	ZVAL_COPY(&elem.data, data);
 	ZVAL_COPY(&elem.priority, priority);
 
+	/* If we know this call came from non inherited SplPriorityQueue it's
+	 * possible to do specialization on the type of the priority parameter. */
+	if (!intern->fptr_cmp) {
+		int type = Z_TYPE(elem.priority);
+		spl_ptr_heap_cmp_func new_cmp =
+			(type == IS_LONG) ? spl_ptr_pqueue_elem_cmp_long :
+			((type == IS_DOUBLE) ? spl_ptr_pqueue_elem_cmp_double : spl_ptr_pqueue_elem_cmp);
+
+		if (intern->heap->count == 0) { /* Specialize empty queue */
+			intern->heap->cmp = new_cmp;
+		} else if (new_cmp != intern->heap->cmp) { /* Despecialize on type conflict. */
+			intern->heap->cmp = spl_ptr_pqueue_elem_cmp;
+		}
+	}
+
 	spl_ptr_heap_insert(intern->heap, &elem, ZEND_THIS);
 
 	RETURN_TRUE;
@@ -820,7 +849,7 @@ PHP_METHOD(SplHeap, top)
 		RETURN_THROWS();
 	}
 
-	ZVAL_COPY_DEREF(return_value, value);
+	RETURN_COPY_DEREF(value);
 }
 /* }}} */
 
@@ -852,10 +881,8 @@ PHP_METHOD(SplMaxHeap, compare)
 
 static void spl_heap_it_dtor(zend_object_iterator *iter) /* {{{ */
 {
-	spl_heap_it *iterator = (spl_heap_it *)iter;
-
 	zend_user_it_invalidate_current(iter);
-	zval_ptr_dtor(&iterator->intern.it.data);
+	zval_ptr_dtor(&iter->data);
 }
 /* }}} */
 
@@ -994,7 +1021,7 @@ PHP_METHOD(SplHeap, current)
 		RETURN_NULL();
 	} else {
 		zval *element = spl_heap_elem(intern->heap, 0);
-		ZVAL_COPY_DEREF(return_value, element);
+		RETURN_COPY_DEREF(element);
 	}
 }
 /* }}} */
@@ -1021,7 +1048,7 @@ PHP_METHOD(SplPriorityQueue, current)
 PHP_METHOD(SplHeap, __debugInfo)
 {
 	if (zend_parse_parameters_none() == FAILURE) {
-		return;
+		RETURN_THROWS();
 	}
 
 	RETURN_ARR(spl_heap_object_get_debug_info(spl_ce_SplHeap, Z_OBJ_P(ZEND_THIS)));
@@ -1031,7 +1058,7 @@ PHP_METHOD(SplHeap, __debugInfo)
 PHP_METHOD(SplPriorityQueue, __debugInfo)
 {
 	if (zend_parse_parameters_none() == FAILURE) {
-		return;
+		RETURN_THROWS();
 	}
 
 	RETURN_ARR(spl_heap_object_get_debug_info(spl_ce_SplPriorityQueue, Z_OBJ_P(ZEND_THIS)));
@@ -1062,97 +1089,76 @@ static const zend_object_iterator_funcs spl_pqueue_it_funcs = {
 
 zend_object_iterator *spl_heap_get_iterator(zend_class_entry *ce, zval *object, int by_ref) /* {{{ */
 {
-	spl_heap_it     *iterator;
-	spl_heap_object *heap_object = Z_SPLHEAP_P(object);
-
 	if (by_ref) {
 		zend_throw_error(NULL, "An iterator cannot be used with foreach by reference");
 		return NULL;
 	}
 
-	iterator = emalloc(sizeof(spl_heap_it));
+	zend_user_iterator *iterator = emalloc(sizeof(zend_user_iterator));
+	zend_iterator_init(&iterator->it);
 
-	zend_iterator_init(&iterator->intern.it);
+	ZVAL_OBJ_COPY(&iterator->it.data, Z_OBJ_P(object));
+	iterator->it.funcs = &spl_heap_it_funcs;
+	iterator->ce       = ce;
+	ZVAL_UNDEF(&iterator->value);
 
-	Z_ADDREF_P(object);
-	ZVAL_OBJ(&iterator->intern.it.data, Z_OBJ_P(object));
-	iterator->intern.it.funcs = &spl_heap_it_funcs;
-	iterator->intern.ce       = ce;
-	iterator->flags           = heap_object->flags;
-	ZVAL_UNDEF(&iterator->intern.value);
-
-	return &iterator->intern.it;
+	return &iterator->it;
 }
 /* }}} */
 
 zend_object_iterator *spl_pqueue_get_iterator(zend_class_entry *ce, zval *object, int by_ref) /* {{{ */
 {
-	spl_heap_it     *iterator;
-	spl_heap_object *heap_object = Z_SPLHEAP_P(object);
-
 	if (by_ref) {
 		zend_throw_error(NULL, "An iterator cannot be used with foreach by reference");
 		return NULL;
 	}
 
-	iterator = emalloc(sizeof(spl_heap_it));
+	zend_user_iterator *iterator = emalloc(sizeof(zend_user_iterator));
+	zend_iterator_init(&iterator->it);
 
-	zend_iterator_init((zend_object_iterator*)iterator);
+	ZVAL_OBJ_COPY(&iterator->it.data, Z_OBJ_P(object));
+	iterator->it.funcs = &spl_pqueue_it_funcs;
+	iterator->ce       = ce;
+	ZVAL_UNDEF(&iterator->value);
 
-	Z_ADDREF_P(object);
-	ZVAL_OBJ(&iterator->intern.it.data, Z_OBJ_P(object));
-	iterator->intern.it.funcs = &spl_pqueue_it_funcs;
-	iterator->intern.ce       = ce;
-	iterator->flags           = heap_object->flags;
-
-	ZVAL_UNDEF(&iterator->intern.value);
-
-	return &iterator->intern.it;
+	return &iterator->it;
 }
 /* }}} */
 
 PHP_MINIT_FUNCTION(spl_heap) /* {{{ */
 {
-	REGISTER_SPL_STD_CLASS_EX(SplHeap, spl_heap_object_new, class_SplHeap_methods);
+	spl_ce_SplHeap = register_class_SplHeap(zend_ce_iterator, zend_ce_countable);
+	spl_ce_SplHeap->create_object = spl_heap_object_new;
+	spl_ce_SplHeap->get_iterator = spl_heap_get_iterator;
+
 	memcpy(&spl_handler_SplHeap, &std_object_handlers, sizeof(zend_object_handlers));
 
 	spl_handler_SplHeap.offset         = XtOffsetOf(spl_heap_object, std);
 	spl_handler_SplHeap.clone_obj      = spl_heap_object_clone;
 	spl_handler_SplHeap.count_elements = spl_heap_object_count_elements;
 	spl_handler_SplHeap.get_gc         = spl_heap_object_get_gc;
-	spl_handler_SplHeap.dtor_obj = zend_objects_destroy_object;
 	spl_handler_SplHeap.free_obj = spl_heap_object_free_storage;
 
-	REGISTER_SPL_IMPLEMENTS(SplHeap, Iterator);
-	REGISTER_SPL_IMPLEMENTS(SplHeap, Countable);
-
-	spl_ce_SplHeap->get_iterator = spl_heap_get_iterator;
-
-	REGISTER_SPL_SUB_CLASS_EX(SplMinHeap,           SplHeap,        spl_heap_object_new, class_SplMinHeap_methods);
-	REGISTER_SPL_SUB_CLASS_EX(SplMaxHeap,           SplHeap,        spl_heap_object_new, class_SplMaxHeap_methods);
-
-	spl_ce_SplMaxHeap->get_iterator = spl_heap_get_iterator;
+	spl_ce_SplMinHeap = register_class_SplMinHeap(spl_ce_SplHeap);
+	spl_ce_SplMinHeap->create_object = spl_heap_object_new;
 	spl_ce_SplMinHeap->get_iterator = spl_heap_get_iterator;
 
-	REGISTER_SPL_STD_CLASS_EX(SplPriorityQueue, spl_heap_object_new, class_SplPriorityQueue_methods);
+	spl_ce_SplMaxHeap = register_class_SplMaxHeap(spl_ce_SplHeap);
+	spl_ce_SplMaxHeap->create_object = spl_heap_object_new;
+	spl_ce_SplMaxHeap->get_iterator = spl_heap_get_iterator;
+
+	spl_ce_SplPriorityQueue = register_class_SplPriorityQueue(zend_ce_iterator, zend_ce_countable);
+	spl_ce_SplPriorityQueue->create_object = spl_heap_object_new;
+	spl_ce_SplPriorityQueue->get_iterator = spl_pqueue_get_iterator;
+
 	memcpy(&spl_handler_SplPriorityQueue, &std_object_handlers, sizeof(zend_object_handlers));
 
 	spl_handler_SplPriorityQueue.offset         = XtOffsetOf(spl_heap_object, std);
 	spl_handler_SplPriorityQueue.clone_obj      = spl_heap_object_clone;
 	spl_handler_SplPriorityQueue.count_elements = spl_heap_object_count_elements;
 	spl_handler_SplPriorityQueue.get_gc         = spl_pqueue_object_get_gc;
-	spl_handler_SplPriorityQueue.dtor_obj = zend_objects_destroy_object;
 	spl_handler_SplPriorityQueue.free_obj = spl_heap_object_free_storage;
 
-	REGISTER_SPL_IMPLEMENTS(SplPriorityQueue, Iterator);
-	REGISTER_SPL_IMPLEMENTS(SplPriorityQueue, Countable);
-
-	spl_ce_SplPriorityQueue->get_iterator = spl_pqueue_get_iterator;
-
-	REGISTER_SPL_CLASS_CONST_LONG(SplPriorityQueue, "EXTR_BOTH",      SPL_PQUEUE_EXTR_BOTH);
-	REGISTER_SPL_CLASS_CONST_LONG(SplPriorityQueue, "EXTR_PRIORITY",  SPL_PQUEUE_EXTR_PRIORITY);
-	REGISTER_SPL_CLASS_CONST_LONG(SplPriorityQueue, "EXTR_DATA",      SPL_PQUEUE_EXTR_DATA);
-
 	return SUCCESS;
 }
 /* }}} */
